import io
import json
import pickle
from abc import ABC
from typing import Tuple, Union, Iterable
from shapely.geometry.base import BaseGeometry
import shapely.affinity as A

from .definition import Shape, Single, Multi
from . import functional as F


Position = Union[str, complex, Tuple[float, float]]


class Base(Shape, ABC):

    def __init__(self, geo: BaseGeometry, reverse: bool = False):
        self._geo = geo
        self._reversed = reverse

    # 这一批运算不改变轮廓类型
    def comp(self) -> None:
        self._reversed = not self._reversed

    def offset(self, vector: Position) -> Shape:
        if isinstance(vector, str):
            assert vector == 'center', '仅支持 "center" 作为入参'  # 平移图像直至中心点为原点
            x, y = self.center
            x = -x
            y = -y
        else:
            x, y = (vector.real, vector.imag) if isinstance(vector, complex) else vector
        self._geo = A.translate(self._geo, x, y)
        return self

    def scale(self, ratio: Union[float, complex, tuple], origin: Position = 0j) -> Shape:
        # 缩放比例支持
        if isinstance(ratio, (float, int)):
            x_fact = y_fact = ratio
        elif isinstance(ratio, complex):
            x_fact = ratio.real
            y_fact = ratio.imag
        else:
            x_fact, y_fact = ratio
        # 原点支持
        if isinstance(origin, complex):
            origin = (origin.real, origin.imag)
        elif isinstance(origin, str):
            assert origin == 'center', '仅支持 "center" 作为入参'
        self._geo = A.scale(self._geo, xfact=x_fact, yfact=y_fact, zfact=0, origin=origin)
        return self

    def rotate(self, degree: float, origin: Union[str, Position] = 0j) -> Shape:
        # 注意注意：此处的 degree 是角度制
        # 原点支持
        if isinstance(origin, complex):
            origin = (origin.real, origin.imag)
        elif isinstance(origin, str):
            assert origin == 'center', '仅支持 "center" 作为入参'
        self._geo = A.rotate(self._geo, angle=degree, origin=origin)
        return self

    def flip_x(self, x0: float) -> Shape:
        self._geo = A.scale(self._geo, xfact=-1, yfact=1, zfact=0, origin=(x0, 0))
        return self

    def flip_y(self, y0: float) -> Shape:
        self._geo = A.scale(self._geo, xfact=1, yfact=-1, zfact=0, origin=(0, y0))
        return self

    def flip(self, degree: float, origin: Position) -> Shape:
        # 直接两次旋转一次对称来做
        if isinstance(origin, str):
            assert origin == 'center', '仅支持 "center" 作为入参'
            x, y = self.center
        else:
            x, y = (origin.real, origin.imag) if isinstance(origin, complex) else origin
        self.rotate(degree=-degree, origin=(x, y))
        self.flip_y(y0=y)
        self.rotate(degree=degree, origin=(x, y))
        return self

    def is_joint(self, other: Shape) -> bool:
        # EMPTY 和 FULL 不属于 Base，不会调用此方法，其判别在自己的逻辑内进行即可
        if other is Shape.EMPTY or other is Shape.FULL: return other.is_joint(self)
        if not self.reversed and not other.reversed:        # 对正常图形来说，直接调库
            return self.geo.intersects(other.geo)
        if self.reversed and other.reversed:                # 两个反形必然相交，因为无穷远处总有一个元素同时属于二者
            return True
        if not self.reversed and other.reversed:            # 一正一反的情况下，需要判断是否反形完全涵盖正形
            return not other.geo.contains(self.geo)
        if self.reversed and not other.reversed:            # 一正一反的情况下，需要判断是否反形完全涵盖正形
            return not self.geo.contains(other.geo)

    def if_contain(self, other: Shape) -> bool:
        # EMPTY 和 FULL 不属于 Base，不会调用此方法，其判别在自己的逻辑内进行即可
        if other is Shape.EMPTY or other is Shape.FULL: return other.is_joint(self)
        if not self.reversed and not other.reversed:        # 对正常图形来说，直接调库
            return self.geo.contains(other.geo)
        if self.reversed and other.reversed:                # 两个反形的包含关系刚好调过来
            return other.geo.contains(self.geo)
        if not self.reversed and other.reversed:            # 正形不可能包含反形
            return False
        if self.reversed and not other.reversed:            # 反形的 geo 与正形完全不相交的情况下即意味着包含
            return not self.geo.intersects(other.geo)

    # 这一批运算会改变轮廓类型
    def inter(self, other: Shape) -> Shape:
        # EMPTY 和 FULL 不属于 Base，不会调用此方法，其判别在自己的逻辑内进行即可
        if other is Shape.EMPTY or other is Shape.FULL: return other.inter(self)
        # 依据 reverse 标志决定真实运算
        if not self.reversed and not other.reversed:        # 正 + 正 -> 正常运算 - 交         A & B = A & B
            return F.inter(self, other, reverse=False)
        if self.reversed and other.reversed:                # 反 + 反 -> 镜像运算 - 反(并)     ~A & ~B = ~(A | B)
            return F.union(self, other, reverse=True)
        if not self.reversed and other.reversed:            # 正 + 反 -> 正斜运算 - 我移除它    A & ~B = A >> B
            return F.remove(self, other, reverse=False)
        if self.reversed and not other.reversed:            # 反 + 正 -> 反斜运算 - 它移除我    ~A & B = B >> A
            return F.remove(other, self, reverse=False)

    def union(self, other: Shape) -> Shape:
        # EMPTY 和 FULL 不属于 Base，不会调用此方法，其判别在自己的逻辑内进行即可
        if other is Shape.EMPTY or other is Shape.FULL:
            return other.union(self)
        # 依据 reverse 标志决定真实运算
        if not self.reversed and not other.reversed:        # 正 + 正 -> 正常运算 - 并             A | B = A | B
            return F.union(self, other, reverse=False)
        if self.reversed and other.reversed:                # 反 + 反 -> 镜像运算 - 反(交)         ~A | ~B = ~(A & B)
            return F.inter(self, other, reverse=True)
        if not self.reversed and other.reversed:            # 正 + 反 -> 正斜运算 - 反(它移除我)    A | ~B = ~(~A & B) = ~(B >> A)
            return F.remove(other, self, reverse=True)
        if self.reversed and not other.reversed:            # 反 + 正 -> 反斜运算 - 反(我移除它)    ~A | B = ~(A & ~B) = ~(A >> B)
            return F.remove(self, other, reverse=True)

    def diff(self, other: Shape) -> Shape:
        # EMPTY 和 FULL 不属于 Base，不会调用此方法，其判别在自己的逻辑内进行即可
        if other is Shape.EMPTY or other is Shape.FULL: return other.diff(self)
        # 依据 reverse 标志决定真实运算
        if not self.reversed and not other.reversed:        # 正 + 正 -> 正常运算 - 异         A ^ B = (A & ~B) | (~A & B) = A ^ B
            return F.diff(self, other, reverse=False)
        if self.reversed and other.reversed:                # 反 + 反 -> 镜像运算 - 异         ~A ^ ~B = (~A & B) | (A & ~B) = A ^ B
            return F.diff(self, other, reverse=False)
        if not self.reversed and other.reversed:            # 正 + 反 -> 正斜运算 - 反(异)      A ^ ~B = (A & B) | (~A & ~B) = (A&B) | ~(A|B) = ~[~(A&B) & (A|B)] = ~[(A|B) >> (A&B)] = ~(A ^ B)
            return F.remove(other, self, reverse=True)
        if self.reversed and not other.reversed:            # 反 + 正 -> 反斜运算 - 反(异)      ~A ^ B = A ^ ~B = ~(A ^ B)
            return F.remove(self, other, reverse=True)

    def remove(self, other: Shape) -> Shape:
        # EMPTY 和 FULL 不属于 Base，不会调用此方法，其判别在自己的逻辑内进行即可
        if other is Shape.EMPTY or other is Shape.FULL: return other.remove(self)
        # 依据 reverse 标志决定真实运算
        if not self.reversed and not other.reversed:        # 正 + 正 -> 正常运算 - 我移除它      A >> B = (A & ~B) = A >> B
            return F.remove(self, other, reverse=False)
        if self.reversed and other.reversed:                # 反 + 反 -> 镜像运算 - 它移除我      ~A >> ~B = (~A & B) = B >> A
            return F.remove(other, self, reverse=False)
        if not self.reversed and other.reversed:            # 正 + 反 -> 正斜运算 - 交           A >> ~B = (A & B) = A & B
            return F.inter(self, other, reverse=False)
        if self.reversed and not other.reversed:            # 反 + 正 -> 反斜运算 - 反(并)       ~A >> B = (~A & ~B) = ~(A|B)
            return F.union(self, other, reverse=True)

    def merge(self, others: Iterable[Shape]) -> Shape:
        # 排除与自身不相交的
        joins = [other for other in others if self.is_joint(other)]
        # 检查有没有 Full, 如果有, 直接返回 Full
        if any((other is Shape.FULL) for other in joins): return Shape.FULL
        # 将自己也加进去
        joins.append(self)
        # 相交的图形将正形反形分离开
        pos_shapes = [other for other in joins if not other.reversed]
        neg_shapes = [other for other in joins if other.reversed]
        # 正形参与 union 运算
        pos_shape = F.union(*pos_shapes, reverse=False)
        # 反形参与 inter 运算
        neg_shape = F.inter(*neg_shapes, reverse=True) if neg_shapes else Shape.EMPTY
        # 二者之和即为所求
        return pos_shape.union(neg_shape)

    def simplify(self, tolerance: float = 0.5) -> Shape:
        geo = self.geo.simplify(tolerance=tolerance, preserve_topology=True)
        return F.norm_multi(geo, reverse=self.reversed)

    def smooth(self, distance: float = 3) -> Shape:
        distance = -distance if self.reversed else distance
        geo = self.geo.buffer(-distance).buffer(distance)
        return F.norm_multi(geo, reverse=self.reversed)

    def buffer(self, distance: float = 3) -> Shape:
        distance = -distance if self.reversed else distance
        geo = self.geo.buffer(distance=distance)
        return F.norm_multi(geo, reverse=self.reversed)

    def standard(self) -> Multi:
        return F.norm_multi(self.geo, reverse=self.reversed)

    @property
    def convex(self) -> Single:
        geo = self._geo.convex_hull
        return Single.SIMPLE(geo=geo)

    @property
    def mini_rect(self) -> Single:
        geo = self._geo.minimum_rotated_rectangle
        return Single.SIMPLE(geo=geo)

    @property
    def region(self) -> Single:
        l, u, r, d = self._geo.bounds
        return Single.REGION(l, u, r, d)

    @property
    def center(self) -> Tuple[int, int]:
        return self._geo.centroid.coords[0]

    @property
    def area(self) -> float:
        return self._geo.area

    @property
    def perimeter(self) -> float:
        return self._geo.length

    @property
    def bounds(self) -> Tuple[int, int, int, int]:
        return self._geo.bounds

    @property
    def geo(self) -> BaseGeometry:
        return self._geo

    @property
    def cls(self) -> type:
        raise NotImplementedError

    def copy(self) -> Shape:
        return self.cls(geo=self._geo, reverse=self.reversed)

    def dumps(self) -> str:
        tp = self.cls.__name__
        rvs = self.reversed
        contour = self.sep_p()
        return f'{tp}\n{rvs}\n{json.dumps(contour)}'

    def dumpb(self, f: io.BufferedWriter):
        tp = self.cls.__name__
        rvs = self.reversed
        contour = self.sep_p()
        pickle.dump((tp, rvs, contour), f)



from typing import List, Tuple, Iterable

import shapely
from shapely.geometry.base import BaseGeometry

from .definition import Shape, Single, Multi, NoParametersException, CoordinatesNotLegalException
from .impl_base import Base
from . import functional as F


class MultiComplexPolygon(Base, Multi):
    """
    多-复连通多边形, 创建方式有五:
    1. 指定 outers
    2. 指定 outers, inners, adjacencies
    3. 指定 geo
    4. 指定 Single 数组
    5. 指定 Multi
    遵循逆序优先规则
    """

    def __init__(
            self,
            outers: List[Tuple[float, float]] = None,
            inners: List[Tuple[float, float]] = None,
            adjacencies: List[int] = None,

            geo: BaseGeometry = None,

            shapes: Iterable[Shape] = None,

            from_p: Tuple[
                List[List[Tuple[float, float]]],    # outers
                List[List[Tuple[float, float]]],    # inners
                List[List[Tuple[float, float]]],    # adjacencies
            ] = None,
            reverse: bool = False,
    ):
        if geo is not None:
            assert isinstance(geo, shapely.MultiPolygon), 'geo 必须是 MultiPolygon'
        elif shapes is not None:
            assert not any(shape.reversed for shape in shapes), '创建 MultiPolygon 时只能采用正形描述'
            geoms = []
            for shape in shapes:
                if not shape: continue
                if isinstance(shape, Single):
                    geoms.append(shape.geo)
                elif isinstance(shape, Multi):
                    geoms.extend(geo for geo in shape.geo.geoms if isinstance(geo, shapely.Polygon))
            geo = shapely.MultiPolygon(polygons=geoms)
        else:
            if from_p is not None:
                outers, inners, adjacencies = from_p
            if outers is None:
                # 没有任何参数的话，要报个错
                raise NoParametersException(f'Any of such parameters have to be provided: (outer, *inners), geo, single, from_p')

            if inners is None and adjacencies is None:
                coords = [(outer, []) for outer in outers]
            else:
                assert inners is not None and adjacencies is not None and len(inners) == len(adjacencies), '孔洞未对齐'
                coords = [
                    (outer, [inner for j, inner in enumerate(inners) if adjacencies[j] == i])
                    for i, outer in enumerate(outers)
                ]
            # 对用户输入进行检查和修复
            geo = shapely.MultiPolygon(polygons=coords)
            geo = F.norm_geo(geo)
            # 创建时要求轮廓必须合法
            if geo is None:
                raise CoordinatesNotLegalException(f'creating single polygon with geo=={type(geo)}')
            elif isinstance(geo, shapely.Polygon):
                geo = shapely.MultiPolygon(polygons=[geo])

        super().__init__(geo=geo, reverse=reverse)

    @property
    def outer(self) -> Multi:
        # 外轮廓(正形)
        outer_geos = [shapely.Polygon(g.exterior.coords) for g in self.geo.geoms]
        geo = shapely.MultiPolygon(outer_geos)
        return Multi.SIMPLE(geo=geo)

    @property
    def inner(self) -> Multi:
        # 内轮廓(负形)
        inner_boundaries = [list(g.interiors) for g in self.geo.geoms]
        inner_boundaries = sum(inner_boundaries, [])
        geos = [shapely.Polygon(b.coords) for b in inner_boundaries]
        geo = shapely.MultiPolygon(polygons=geos)
        return Multi.SIMPLE(geo=geo)

    def sep_in(self) -> Tuple[Multi, Multi]:
        # 内分解
        return self.outer, self.inner

    def sep_out(self) -> List[Single]:
        # 外分解
        return [Single.COMPLEX(geo=g) for g in self.geo.geoms]

    def sep_p(self) -> Tuple[
        List[List[Tuple[int, int]]],
        List[List[Tuple[int, int]]],
        List[int]
    ]:
        # 点分解
        singles = [Single.COMPLEX(geo=g) for g in self.geo.geoms]
        outers = []
        inners = []
        adjacencies = []
        for i, single in enumerate(singles):
            p, qs = single.sep_p()
            outers.append(p)
            inners.extend(qs)
            adjacencies.extend([i] * len(qs))
        return outers, inners, adjacencies

    @property
    def cls(self) -> type:
        return MultiComplexPolygon


Multi.COMPLEX = MultiComplexPolygon
